const { app, BrowserWindow, ipcMain, dialog, Menu } = require('electron');
const Store = require("electron-store");
const fs = require("fs");
const path = require("path");
const util = require('util')
let Validator = require("jsonschema").Validator;
let validator = new Validator();

let all_schemas = [
  path.join(app.getAppPath(), "src/schemas/current_schema.json"),
  path.join(app.getAppPath(), "src/schemas/1.0.0_schema.json"),
]
let schema_updaters = {
  "1.0.0_schema.json": update_1_0_0_schema
}
let jsdom = require("jsdom");
const { JSDOM } = jsdom;
const { window } = new JSDOM();
let $ = jQuery = require('jquery')(window);

const store = new Store();
let current_file = null;
let menu_template = [
  {
    label: "File",
    submenu: [
      {
        label: "Save",
        accelerator: "Ctrl+S",
        click: function () {
          let window = BrowserWindow.getFocusedWindow();
          window.webContents.send("save", false);
        }
      },
      {
        label: "Save As",
        accelerator: "Ctrl+Shift+S",
        click: function () {
          let window = BrowserWindow.getFocusedWindow();
          window.webContents.send("save", true);
        }
      },
      {
        label: "Open",
        accelerator: "Ctrl+O",
        click: function () {
          openFile();
        }
      },
      {
        label: "Open Recent",
        submenu:[]
      },
      {
        label: "Exit",
        click: function () {
          app.quit();
        }
      }
    ]
  },
  {
    label: "Edit",
    submenu: [
      {
        label: "Copy",
        accelerator: "Ctrl+C",
        click: function () {
          let window = BrowserWindow.getFocusedWindow();
          window.webContents.send("copy-to-clipboard");
        }
      },
    ]
  }
]

// Handle creating/removing shortcuts on Windows when installing/uninstalling.
if (require('electron-squirrel-startup')) { // eslint-disable-line global-require
  app.quit();
}

// This method will be called when Electron has finished
// initialization and is ready to create browser windows.
// Some APIs can only be used after this event occurs.
app.on('ready', createSplashScreen);

// Quit when all windows are closed, except on macOS. There, it's common
// for applications and their menu bar to stay active until the user quits
// explicitly with Cmd + Q.
app.on('window-all-closed', () => {
  if (process.platform !== 'darwin') {
    app.quit();
  }
});

app.on('activate', () => {
  // On OS X it's common to re-create a window in the app when the
  // dock icon is clicked and there are no other windows open.
  if (BrowserWindow.getAllWindows().length === 0) {
    createSplashScreen();
  }
});

function createSplashScreen() {
  const splash_screen = new BrowserWindow({
    width: 806,
    height: 529,
    autoHideMenuBar: true,
    webPreferences: {
      nodeIntegration: true
    },
    resizable: false,
    frame: true
  });

  if (!store.has("recent_files")) {
    store.set("recent_files", []);
  } else {
    updateRecentFiles();
  }

  splash_screen.loadFile(path.join(app.getAppPath(), "src/splash.html"));
  // Open the DevTools.
  // splash_screen.webContents.openDevTools();
  // store.openInEditor();
}

function createEditorWindow(json) {
  let main_window = new BrowserWindow({
    width: 1000,
    height: 600,
    autoHideMenuBar: false,
    webPreferences: {
      nodeIntegration: true
    }
  })
  main_window.maximize();
  updateMenuBar();

  main_window.loadFile(path.join(app.getAppPath(), "src/editor.html"))
    .then(() => {
      main_window.webContents.send("open", json);
    });
  // Open the DevTools.
  // main_window.webContents.openDevTools();
  return main_window;
}

function createNew() {
  let old_window = BrowserWindow.getFocusedWindow();
  let json = loadJson(path.join(app.getAppPath(), "src/default.json"));
  let main_window = createEditorWindow(json);
  old_window.close();
}

function chooseFile(title, for_save=false) {
  let options = {
    title: title,
    defaultPath: store.get("default_dir", app.getPath("documents")),
    filters: [
      {name: 'json', extensions: ['json']},
      {name: 'All Files', extensions: ['*']}
    ]
  }

  let file_path;
  if (for_save) {
    file_path = dialog.showSaveDialogSync(options);
  } else {
    file_path = dialog.showOpenDialogSync(options);
  }
  if (file_path) {
    if (for_save) {
      return file_path
    } else {
      return file_path[0];
    }
  }
}

function openFile(file_path="", new_file=false) {
  if (file_path === "") {
    file_path = chooseFile("Open website data");
  }
  if (file_path && fs.existsSync(file_path)) {
    let result = checkFile(file_path);
    if (result) {
      let json = getUpdatedJson(result);
      updateRecentFiles(file_path);
      current_file = file_path
      if (new_file) {
        let old_window = BrowserWindow.getFocusedWindow();
        createEditorWindow(json);
        old_window.close();
      } else {
        let window = BrowserWindow.getFocusedWindow();
        window.webContents.send("open", json);
      }
    }
  }
}

function saveFile(json_data, create_new_file) {
  if (current_file == null || create_new_file) {
    let file_path = chooseFile("Save website data", true);
    if (file_path) {
      updateRecentFiles(file_path);
      current_file = file_path;
    }
  }
  if (current_file) {
    fs.writeFileSync(current_file, JSON.stringify(json_data, null, 4));
  }
}

function updateRecentFiles(new_file=null) {
  let recent_files = store.get("recent_files");
  if (new_file != null) {
    store.set("default_dir", new_file);
    if (recent_files.includes(new_file)) {
      recent_files.splice(recent_files.indexOf(new_file), 1);
    }
    recent_files.splice(0, 0, new_file);
  }
  let files_to_delete = [];
  $.each(recent_files, (i, file) => {
    if (!fs.existsSync(file)) {
      console.log(file);
      files_to_delete.push(i);
    }
  });
  for (let index of files_to_delete) {
    recent_files.splice(index, 1);
  }
  store.set("recent_files", recent_files);
  updateMenuBar();
}

function updateMenuBar() {
  let recent_files = store.get("recent_files");
  for (let menu of menu_template[0].submenu) {
    if (menu.label === "Open Recent") {
      menu.submenu = []
      for (let file of recent_files) {
        if (fs.existsSync(file)) {
          menu.submenu.push({
            label: file, click: () => {
              openFile(file);
            }
          });
        }
      }
    }
  }
  let menu = Menu.buildFromTemplate(menu_template);
  Menu.setApplicationMenu(menu);
}

function loadJson(file_path) {
  let file = fs.readFileSync(file_path);
  return JSON.parse(file.toString());
}

function checkFile(file_path) {
  let json = loadJson(file_path);
  let result;
  let valid_result = false;
  for (let schema_path of all_schemas) {
    let schema = loadJson(schema_path);
    result = validator.validate(json, schema);
    if (result.valid) {
      let split_schema = schema_path.split(/[\\/]/);
      result = {"schema": split_schema[split_schema.length-1], "result": result};
      valid_result = true;
      break;
    }
  }
  if (valid_result) {
    return result;
  }
}

function getUpdatedJson(validation_result) {
  let json = validation_result["result"].instance;
  let schema = validation_result["schema"];
  if (schema === "current_schema.json") {
    return json
  }
  return schema_updaters[schema](json);
}

function update_1_0_0_schema(json) {
  $.each(json, (i, table_data) => {
    table_data["id"] = "table"+i;
    table_data["tab_name"] = table_data["tab-name"];
    delete table_data["tab-name"];
    let col_dict = {};
    $.each(table_data["col-def"], (j, col_data) => {
      col_dict[col_data["field"]] = "column"+j;
      col_data["field"] = "column"+j;
    });
    table_data["columns"] = table_data["col-def"];
    delete table_data["col-def"];
    $.each(table_data["rows"], (j, row_data) => {
      row_data["uid"] = "row"+j;
      $.each(col_dict, (old_key, new_key) => {
        row_data[new_key] = row_data[old_key];
        delete row_data[old_key];
      });
    });
  });
  json = {
    "version": "1.1.0",
    "tables": json,
  }
  return json;
}

function duplicateJson(json) {
  return JSON.parse(JSON.stringify(json));
}

function removeRecentFile(file) {
  let recent_files = store.get("recent_files");
  recent_files.splice(recent_files.indexOf(file), 1);
  store.set("recent_files", recent_files);
}

ipcMain.handle("get-store-value", (event, key) => {
  return store.get(key)
})

ipcMain.handle("remove-recent-file", (event, file) => {
  removeRecentFile(file);
})

ipcMain.on("create-new", (event) => {
  createNew();
})

ipcMain.on("save", (event, json_data, create_new_file) => {
  saveFile(json_data, create_new_file);
})

ipcMain.on("open", (event, file_path="") => {
  openFile(file_path, true)
})
